#include "plugin.h"
#include <errno.h>
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <strings.h>


extern int DEBUG;
extern int errno;
// Linked List methods
// This is how the playlist is going to be managed

void initialize(Node * l)
{
    l = NULL;
}

Node *newNode()
{
    Node *n = (Node *) NPN_MemAlloc(sizeof(Node));
    n->url[0] = '\0';
    n->fname[0] = '\0';
    n->next = NULL;
    n->bytes = 0;
    n->play = 1;
    n->playlist = 0;
    n->speed = -1;
    n->mmsstream = 0;
    n->remove = 1;
    n->status = STATE_NEWINSTANCE;
    n->retrieved = 0;
    n->cancelled = 0;
    n->bytes = 0;
    n->totalbytes = 0;
    n->cachebytes = 0;
    n->actual_x = 0;
    n->actual_y = 0;
    n->play_x = 0;
    n->play_y = 0;
    n->next = NULL;
    n->area = NULL;	/* linked list for smil <area> tags */
    return n;
}

void deleteNode(Node * n)
{
    if (n != NULL) {
	n->url[0] = '\0';
	n->fname[0] = '\0';
	n->bytes = 0;
	n->play = 1;
	n->mmsstream = 0;
	n->remove = 1;
	n->status = STATE_NEWINSTANCE;
	n->retrieved = 0;
	NPN_MemFree(n);
	n = NULL;
    }
}

void copyNode(Node * dest, Node * src)
{
    if (src != NULL && dest != NULL) {
	strlcpy(dest->url, src->url, sizeof(dest->url));
	strlcpy(dest->fname, src->fname, sizeof(src->fname));
	dest->bytes = src->bytes;
	dest->play = src->play;
	dest->mmsstream = src->mmsstream;
	dest->playlist = src->playlist;
	dest->speed = dest->speed;
	dest->remove = src->remove;
	dest->cancelled = src->cancelled;
	dest->retrieved = dest->retrieved;
	dest->area = src->area;
    }
}

void deleteList(Node * l)
{
    Node *oldList;

    if (DEBUG)
	printf("Entering deleteList\n");
    while (l != NULL) {
	oldList = l;
	if (DEBUG) {
	    printf("DELETE %s\n", l->url);
	    if (l->remove)
		printf("File delete %s\n", l->fname);
	    printf("size: %i\n", l->bytes);
	    printf("play: %i\n", l->play);
	    printf("playlist: %i\n", l->playlist);
	    printf("speed: %i\n", l->speed);
	    printf("cancelled: %i\n", l->cancelled);
	    printf("retrieved: %i\n", l->retrieved);
	    printf("mms stream: %i\n", l->mmsstream);
	    printf("cache bytes: %li\n", l->cachebytes);
	    printf("Actual Size: %ix%i\n", l->actual_x, l->actual_y);
	    printf("Play Size: %ix%i\n", l->play_x, l->play_y);

	}
	if ((l->remove == 1) || (l->remove == 0 && l->play == 0)) {
	    if (strlen(l->fname) != 0) {
		remove(l->fname);
	    }
	}
	l = l->next;
	deleteNode(oldList);
    }

}

/* Inserts node newnode after node item */
void insertafter(Node * item, Node * newnode)
{
    if (item == NULL || newnode == NULL) {
	if (DEBUG)
	    printf("Cannot insert\n");
    } else {
	newnode->next = item->next;
	item->next = newnode;
    }
}

void addToEnd(Node * l, Node * newnode)
{

    if (l == NULL) {
	l = newnode;
    } else {
	while (l->next != NULL)
	    l = l->next;
	l->next = newnode;
    }

}

void addToList(nsPluginInstance * instance, char *item, Node * parent,
	       int speed)
{
    Node *node = NULL;
    Node *n;
    char tmpdir[1024];
    char *tmp;
    char *filename;
    char localitem[1024];


    if (!isMms(item)) {
	if (strncasecmp(item, "http", 4) != 0) {

	    if (item[0] != '/') {
		strlcpy(tmpdir, item, sizeof(tmpdir));	// reuse the buffer
		if (instance->baseurl != NULL) {
		    strlcpy(localitem, instance->baseurl, 1024);
		} else {
//                  if (This->url != NULL)
//                      This->baseurl = getURLBase(This->url);
//                  if (This->href != NULL)
//                      This->baseurl = getURLBase(This->href);
//                  strlcpy(localitem, This->baseurl, 1024);
		}
		strlcat(localitem, tmpdir, 1024);
	    } else {
		if (instance->hostname != NULL) {
		    snprintf(tmpdir, sizeof(tmpdir), "http://%s/%s",
			     instance->hostname, item);
		    strlcpy(localitem, tmpdir, 1024);
		}
	    }

	} else {
	    // if :80 is in the URL, cut it out
	    strlcpy(localitem, item, sizeof(localitem));
	    tmp = strstr(localitem, ":80");
	    if (tmp != NULL) {
		*tmp = '\0';
		strlcat(localitem, tmp + 3, 1024);
	    }
	}
    } else {
	strlcpy(localitem, item, 1024);
    }


    n = instance->td->list;
    if (DEBUG)
	printf("BUILD - traversing playlist\n");
    while (n != NULL) {
	if (URLcmp(n->url, localitem) == 0) {
	    if (DEBUG)
		printf("found a match, possible self reference stream\n");
	    break;
	}
	n = n->next;
    }

    if (n == NULL) {
	// add it in
	if (DEBUG)
	    printf("adding url %s\n", localitem);
	node = newNode();
	strlcpy(node->url, localitem, sizeof(node->url));
	node->speed = speed;
	addToEnd(parent, node);
	parent->play = 0;
	if (isMms(item)) {
	    node->mmsstream = 1;
	} else {
	    filename = getURLFilename(localitem);
	    if (instance->keep_download == 1 && filename != NULL) {
		snprintf(node->fname, sizeof(node->fname), "%s/%s",
			 instance->download_dir, filename);
		if (DEBUG)
		    printf("Assigned name: %s\n", node->fname);
		node->remove = 0;
	    }
	    if (filename != NULL)
		NPN_MemFree(filename);
	    NPN_GetURL(instance->mInstance, node->url, NULL);
	}
    } else {
	if (n->playlist == 1)
	    n->mmsstream = 1;
    }
}


void insert_area(Node * parent, char *tagtarget, char *tagurl, int tagtime)
{
	area *runner;
	area *n = (area *) NPN_MemAlloc(sizeof(area));
	
	strcpy(n->url, tagurl);
	strcpy(n->target, tagtarget);
	n->begin = tagtime;
	n->next = NULL;

	while( parent->next != NULL )
		parent=parent->next;
	
	if( parent->area == NULL ){
		parent->area = n;
		return;
	}

	runner = parent->area;
	
	while( runner->next != NULL )
		runner = runner->next;
	
	runner->next = n;
}

void find_area_tags( const char *smilbuffer, Node *parent )
{
	char *startarea;
	char *endvideo;
	char *start;

	int tagtime=0;
	char tagtarget[128];
	char tagurl[1024];

	endvideo = strstr( smilbuffer, "</video");
	startarea = strstr( smilbuffer, "<area" );
	
	while ( startarea && startarea < endvideo ){
		start = strstr( startarea + 5, "begin=" );
		if( start ){
			start += 6;
			if (*start=='"')
				++start;
			tagtime = atoi(start);
		}
		start = strstr( startarea + 5, "target=" );
		if( start ){
			start += 7;
			if (*start=='"')
				++start;
			sscanf( start, "%127[^\" \t]", tagtarget );
		}
		start = strstr( startarea + 5, "href=" );
		if( start ){
			start += 5;
			if (*start=='"')
				++start;
			sscanf( start, "%1024[^\" \t]", tagurl );
		}
		
		insert_area(parent, tagtarget, tagurl, tagtime);
		startarea = strstr( startarea + 5, "<area" );
	}
		
}

void buildPlaylist(nsPluginInstance * instance, char *file, Node * parent)
{
    FILE *fp;
    char buffer[16*1024];	/* We read this much chars from the file? */
    char buffer_lower[16*1024];	/* this may not be enough for smil files... (pipas)*/
    char url[1024];
    int size, found;
    char *p;
    char *nextrmda = NULL;
    char *rdrf;
    char *rmdr;
    char *sub, *refurl;
    int i;
    int remainder;
    char *seqtag, *rate, *videotag, *videosrc;
    unsigned int code;
    int speed;
    Node *n;
    int speed_low, speed_med, speed_high;
    unsigned char *c1, *c2, *c3, *c4;

    fp = fopen(file, "r");

    found = 0;
    if (fp != NULL) {
	while (!feof(fp)) {
	    memset(buffer, 0, sizeof(buffer));
	    memset(buffer_lower, 0, sizeof(buffer));
	    memset(url, 0, sizeof(url));

	    size = fread(buffer, 1, sizeof(buffer) - 1, fp);
	    buffer[size] = '\0';

	    if (DEBUG)
		printf("READ: %i\n%s\n", size, buffer);

	    // quicktime files have blocks of 4 characters..
	    // rdrf,XXXX,url ,XXXX,url ending with 00

	    if (memmem(buffer, size, "rmda", 4)) {
		if (DEBUG)
		    printf("found QT\n");

		p = buffer;
		// find the first data element
		p = (char *) memmem(p, size, "rmda", 4);
		if (size > 4) {
		    if (p != NULL)
			p = p + 4;
		    nextrmda = (char *) memmem(p, size - 4, "rmda", 4);
		    if (nextrmda == NULL)
			nextrmda = buffer + size;
		}
		refurl = NULL;
		speed = -1;
		code = 0;

		while ((p != NULL) && ((long) p < ((long) buffer + size))) {
		    rdrf =
			(char *) memmem(p,
					size - ((long) nextrmda -
						(long) p), "rdrf", 4);
		    rmdr =
			(char *) memmem(p,
					size - ((long) nextrmda -
						(long) p), "rmdr", 4);

		    if (rdrf != NULL) {
			code = (unsigned int) (rdrf[15]);
			rdrf = rdrf + 16;
			refurl = (char *) rdrf;
			if (DEBUG) {
			    printf("URL: %s\n", (refurl));
			    printf("CODE: %d\n", code);
			}
		    }

		    if (rmdr != NULL) {
			rmdr = rmdr + 8;

			c1 = (unsigned char *) rmdr;
			c2 = (unsigned char *) rmdr + 1;
			c3 = (unsigned char *) rmdr + 2;
			c4 = (unsigned char *) rmdr + 3;
			speed =
			    ((unsigned int) c1[0] << 24) +
			    ((unsigned int) c2[0] << 16) +
			    ((unsigned int) c3[0] << 8) +
			    (unsigned int) c4[0];

			if (parent->speed > speed)
			    speed = parent->speed;
		    }
		    if (refurl != NULL) {
			if (code == 163 || code == 165 || code == 167
			    || code == (unsigned int) -93
			    || code == (unsigned int) -91
			    || code == (unsigned int) -89) {
			    if (DEBUG)
				printf("Skipped URL: %s\n", refurl);
			} else {
			    parent->playlist = 1;
			    addToList(instance, refurl, parent, speed);
			    printf("ADDED URL: %s\n", (refurl));
			    printf("code: %i\n speed %i\n", code, speed);
			    found = 1;
			    refurl = NULL;
			    speed = -1;
			    code = 0;
			}
		    }
		    if ((nextrmda - size) == p)
			break;
		    p = nextrmda + 4;
		    nextrmda =
			(char *) memmem(p,
					size - ((long) p - (long) buffer),
					"rmda", 4);
		    if (nextrmda == NULL)
			nextrmda = buffer + size;

		}

		break;		// End QT Reference file
	    }
	    // smil format detection
	    snprintf((char *) buffer_lower, sizeof(buffer_lower), "%s", buffer);
	    lowercase((char *) buffer_lower);
	    p = buffer_lower;
	    if (memmem(p, size, "<smil", 5)) {
		while (p != NULL) {
		    speed = -1;
		    remainder = size - ((long) p - (long) buffer_lower);
		    if (remainder < 1)
			break;
		    // seq tags are optional, but if we do have
		    // one see if it has a speed and if so, find it
		    seqtag = (char *) memmem(p, remainder, "<seq", 4);
		    if (seqtag != NULL) {
			if (DEBUG)
			    printf("[smil] found seq tag\n");
			remainder =
			    size - ((long) seqtag - (long) buffer_lower);
			rate =
			    (char *) memmem(seqtag, remainder,
					    "system-bitrate", 14);
			if (rate != NULL) {
			    rate = strchr(rate, '\"');
			    if (rate != NULL)
				rate++;
			    sscanf(rate, "%i", &speed);
			    if (DEBUG)
				printf("[smil] found bit rate %i\n",
				       speed);
			} else {
			    // if we don't fine a speed, then set the speed to -1
			    speed = -1;
			}
		    } else {
			// if we didn't find a seqtag, then set seqtag equal to p
			// so that the next block will work
			seqtag = (char *) p;
		    }
		    // we have a speed now, lets find all the src
		    // files in the seq block
		    videotag =
			(char *) memmem(seqtag, remainder, "<video", 6);
		    while (videotag != NULL) {
			if (DEBUG)
			    printf("[smil] found video tag\n");
			remainder =
			    size - ((long) videotag - (long) buffer_lower);
			videosrc =
			    (char *) memmem(videotag, remainder, "src", 3);
			if (videosrc != NULL) {
			    videosrc = strchr(videosrc, '\"');
			    if (videosrc == NULL)
				break;
			    i = (long) videosrc - (long) buffer_lower;
			    sscanf(buffer + i, "\"%1023[^\"]\"", url);
			    if (DEBUG)
				printf("[smil] found url %s\n", url);
			    parent->playlist = 1;
			    addToList(instance, url, parent, speed);
			    found = 1;
			}

			find_area_tags( videosrc, instance->td->list );

			videotag =
			    (char *) memmem(videosrc, remainder, "<video",
					    6);
			seqtag =
			    (char *) memmem(videosrc, remainder, "<seq",
					    4);
			if (videotag == NULL)
			    break;
			if (seqtag == NULL && videotag != NULL)
			    continue;
			if ((long) videotag > (long) seqtag)
			    videotag = NULL;
		    }
		    p = (char *) seqtag;
		}
	    }
	    // some mp3 files start with ID3, not a reference file
	    // not all MP3 files have the ID tag
	    if (strncasecmp(buffer, "ID3", 3) == 0) {
		if (DEBUG)
		    printf("found MP3\n");
		break;
	    }
	    // asx files are XMLish
	    // <Ref href = "url" />
	    if (strncasecmp(buffer, "<asx", 4) == 0) {
		if (DEBUG)
		    printf("found ASX file\n");
		snprintf(buffer_lower, 4096, "%s", buffer);
		lowercase(buffer_lower);
		p = buffer_lower;
		while (p != NULL) {
		    i = size - ((long) buffer_lower - (long) p);
		    if (i < 1)
			break;
		    p = (char *) memmem(p, i, "href", 4);
		    if (p == NULL)
			break;
		    p = p + 4;
		    p = strchr(p, '\"') + 1;
		    if (p != NULL) {
			i = (long) p - (long) buffer_lower;
			if (i >= size)
			    break;
			snprintf(url, 1024, "%s", buffer + i);
		    }
		    sub = strchr(url, '\"');
		    if (sub != NULL) {
			*sub = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			addToList(instance, url, parent, -1);
			p = p + strlen(url) + 1;
			found = 1;
		    } else {
			break;
		    }
		}
		break;
	    }
	    // asf files
	    if (strncasecmp(buffer, "[reference]", 11) == 0) {
		if (DEBUG)
		    printf("found ASF file\n");
		snprintf(buffer_lower, 4096, "%s", buffer);
		lowercase(buffer_lower);
		p = buffer_lower;
		while (p != NULL) {
		    i = size - ((long) buffer_lower - (long) p);
		    if (i < 1)
			break;
		    p = (char *) memmem(p, i, "ref", 3);
		    if (p == NULL)
			break;
		    p = p + 3;
		    p = strchr(p, '=') + 1;
		    if (p != NULL) {
			i = (long) p - (long) buffer_lower;
			snprintf(url, 1024, "%s", buffer + i);
		    }
		    sub = strchr(url, '\r');
		    if (sub == NULL)
			sub = strchr(url, '\n');
		    if (sub == NULL)
			sub = strchr(url, EOF);
		    if (sub != NULL) {
			*sub = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			parent->playlist = 1;
			addToList(instance, url, parent, -1);
			p = p + strlen(url) + 1;
			found = 1;
		    }
		}
		break;
	    }

	    if (strncasecmp(buffer, "rtsp://", 7) == 0) {
		p = buffer;
		while (p != NULL) {
		    i = size - ((long) p - (long) buffer);
		    p = (char *) memmem(p, i, "rtsp://", 7);
		    if (p == NULL) {
			break;
		    } else {
			snprintf(url, 1024, "%s", p);
			sub = strchr(url, 0xd);
			if (sub != NULL)
			    sub[0] = '\0';
			sub = strchr(url, 0xa);
			if (sub != NULL)
			    sub[0] = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			parent->playlist = 1;
			addToList(instance, url, parent, -1);
			p = p + strlen(url) + 1;
			found = 1;
		    }
		}
	    }

	    if (strncasecmp(buffer, "http://", 7) == 0) {
		p = buffer;
		while (p != NULL) {
		    i = size - ((long) p - (long) buffer);
		    p = (char *) memmem(p, i, "http://", 7);
		    if (p == NULL) {
			break;
		    } else {
			snprintf(url, 1024, "%s", p);
			sub = strchr(url, 0xd);
			if (sub != NULL)
			    sub[0] = '\0';
			sub = strchr(url, 0xa);
			if (sub != NULL)
			    sub[0] = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			parent->playlist = 1;
			addToList(instance, url, parent, -1);
			p = p + strlen(url) + 1;
			found = 1;
		    }
		}
	    }


	    if (!found)
		break;
	}
	fclose(fp);

	speed_low = 0;
	speed_med = 0;
	speed_high = 0;

	if (DEBUG)
	    printf("Scanning for speed\n");
	n = instance->td->list;
	while (n != NULL) {
	    if ((n->speed > 0) && (n->play == 1)) {
		if (speed_low == 0) {
		    speed_low = n->speed;
		    speed_med = n->speed;
		    speed_high = n->speed;
		}

		if ((speed_low == speed_med) && (n->speed > speed_med))
		    speed_med = n->speed;

		if (n->speed < speed_low)
		    speed_low = n->speed;

		if (n->speed > speed_high)
		    speed_high = n->speed;

		if ((n->speed > speed_med) && (n->speed < speed_high))
		    speed_med = n->speed;

		if ((n->speed < speed_high) && (n->speed >= speed_low)
		    && (speed_high == speed_med))
		    speed_med = n->speed;
	    }
	    n = n->next;
	}
	if (DEBUG)
	    printf("low = %i, med = %i, high = %i\n", speed_low, speed_med,
		   speed_high);

	n = instance->td->list;
	while (n != NULL) {
	    if ((n->speed > 0) && (n->play == 1)) {
		if (instance->qt_speed == SPEED_LOW) {
		    if (n->speed != speed_low) {
			n->play = 0;
			n->cancelled = 1;
		    }
		}

		if (instance->qt_speed == SPEED_MED) {
		    if (n->speed != speed_med) {
			n->play = 0;
			n->cancelled = 1;
		    }
		}

		if (instance->qt_speed == SPEED_HIGH) {
		    if (n->speed != speed_high) {
			n->play = 0;
			n->cancelled = 1;
		    }
		}
	    }
	    n = n->next;
	}
	if (DEBUG)
	    printf("playlist adjustment complete\n");

    } else {
	printf("Build Playlist error opening file %i : %s\n", errno,
	       strerror(errno));
    }
}
