#include "plugin.h"
#include "../config.h"
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>

#define STATE_RESET 0
#define STATE_NEW 1
#define STATE_HAVEURL 3
#define STATE_WINDOWSET 4
#define STATE_READY 5
#define STATE_QUEUED 6
#define STATE_DOWNLOADING 7
#define STATE_DOWNLOADED_ENOUGH 8
#define STATE_CANCELLED 11

#define STATE_NEWINSTANCE 100
#define STATE_GETTING_PLAYLIST 110
#define STATE_STARTED_PLAYER 115
#define STATE_PLAYLIST_COMPLETE 120
#define STATE_PLAYLIST_NEXT 125
#define STATE_PLAYING 130
#define STATE_PLAY_COMPLETE 140
#define STATE_PLAY_CANCELLED 150

// speed options
#define SPEED_LOW 1
#define SPEED_MED 2
#define SPEED_HIGH 3

// JavaScript Playstates
#define JS_STATE_UNDEFINED 	0
#define JS_STATE_STOPPED 	1
#define JS_STATE_PAUSED 	2
#define JS_STATE_PLAYING	3
#define JS_STATE_SCANFORWARD	4
#define JS_STATE_SCANREVERSE 	5
#define JS_STATE_BUFFERING	6
#define JS_STATE_WAITING	7
#define JS_STATE_MEDIAENDED	8
#define JS_STATE_TRANSITIONING  9
#define JS_STATE_READY		10
#define JS_STATE_RECONNECTING	11


#ifndef STRUCTURES
#define STRUCTURES

typedef struct area {
	char url[1024];		/* url in smil <area> tag */
	char target[128];	/* target in <area>	  */
	int begin;		/* begin (seconds)	  */
	struct area *next;	/* this is linked list    */
} area;

typedef struct lnode {
    char url[1024];
    char fname[1024];
    int status;
    int retrieved;
    int play;
    int speed;
    int playlist;
    int mmsstream;
    int remove;
    int cancelled;
    long int bytes;
    long int totalbytes;
    long int cachebytes;
    int actual_x, actual_y;
    int play_x, play_y;
    struct area *area;
    struct lnode *next;
} Node;

typedef struct _ThreadData {
    Widget w;
    nsPluginInstance *instance;
    char argv[50][1024];
    Node *list;
} ThreadData;

#endif

// function defintions, grouped by source file
#ifndef HAVE_MEMMEM
extern void *memmem(const void *haystack, const size_t haystack_len,
		    const void *needle, const size_t needle_len);
#endif

#ifndef HAVE_STRLCPY
extern size_t strlcpy(char *dst, const char *src, size_t siz);
#endif

#ifndef HAVE_STRLCAT
extern size_t strlcat(char *dst, const char *src, size_t siz);
#endif

// plugin-setup.cpp
char *GetMIMEDescription();
NPError GetValue(NPPVariable variable, void *value);
void New(nsPluginInstance * instance, nsPluginCreateData * parameters);
void LoadConfigFile(nsPluginInstance * instance);

// plugin-support.cpp
void lowercase(char string[]);
int fexists(char *file);
char *getURLBase(char *url);
char *getURLHostname(char *url);
char *getURLFilename(char *url);
int isMms(char *url);
void sendCommand(FILE * control, char *command);
int URLcmp(const char *url1, const char *url2);
extern void remove_quotes(char *url);

//plugin-list.cpp
void initialize(Node * l);
Node *newNode();
void deleteNode(Node * n);
void copyNode(Node * dest, Node * src);
void deleteList(Node * l);
void insertafter(Node * item, Node * newnode);
void addToEnd(Node * l, Node * newnode);
void buildPlaylist(nsPluginInstance * instance, char *file, Node * parent);

//plugin-ui.cpp
void DrawUI(Widget w, nsPluginInstance * instance, char *message,
	    int FullRedraw, int percent);
extern void RedrawCB(Widget widget, XtPointer client_data,
		     XtPointer call_data);
extern void DestroyCB(Widget widget, XtPointer client_data,
		      XtPointer call_data);

//plugin-threads.cpp
void SetupPlayer(nsPluginInstance * instance, XEvent * event);
void *playPlaylist(void *td);
