#include "plugin.h"
#include <strings.h>

extern int DEBUG;

void lowercase(char string[])
{
    int i = 0;

    while (string[i]) {
	string[i] = tolower(string[i]);
	i++;
    }

    return;
}

int fexists(char *file)
{

    FILE *fp;

    fp = fopen(file, "r");
    if (fp != NULL) {
	fclose(fp);
	return 1;
    } else {
	return 0;
    }

}

char *getURLHostname(char *url)
{

    char *hostname;
    char *loc;
    int i, len;

    if (DEBUG)
	printf("entering getURLHostname\n");
    if (url == NULL)
	return NULL;
    len = strlen(url);
    if (len == 0)
	return NULL;
    hostname = (char *) NPN_MemAlloc(sizeof(char) * (len + 1));
    strcpy(hostname, url);

    loc = strstr(url, "://");
    if (loc == NULL) {
	NPN_MemFree(hostname);
	return NULL;
    }
    loc = loc + 3;

    i = 0;

    while (loc[0] != '/') {
	hostname[i] = loc[0];
	loc++;
	i++;
	if (i > len) {
	    i = 0;
	    break;
	}
    }
    hostname[i] = '\0';
    if (DEBUG)
	printf("exiting getURLHostname with %s\n", hostname);

    return hostname;

}

char *getURLFilename(char *url)
{

    char *filename;
    char *tmp;
    int len;

    if (url == NULL)
	return NULL;
    len = strlen(url);
    if (len == 0)
	return NULL;
    filename = (char *) NPN_MemAlloc(sizeof(char) * (len + 1));
    tmp = rindex(url, '/');
    if (tmp == NULL) {
	strcpy(filename, url);
	return filename;
    }
    strcpy(filename, tmp);
    return filename;
}

char *getURLBase(char *url)
{

    char *base;
    int i, len;

    if (DEBUG)
	printf("in getURLBase url: %s\n", url);
    if (url == NULL)
	return NULL;
    len = strlen(url);
    if (len == 0)
	return NULL;
    base = (char *) NPN_MemAlloc(sizeof(char) * (len + 1));
    strcpy(base, url);
    if (DEBUG)
	printf("in getURLBase base: %s\n", base);

    for (i = len - 1; i >= 0; i--) {
	if (base[i] != '/') {
	    base[i] = '\0';
	} else {
	    break;
	}
    }
    if (strlen(base) == 0) {
	NPN_MemFree(base);
	base = NULL;
    }
    if (base != NULL) {
	if (DEBUG)
	    printf("exiting URL base with %s\n", base);
    } else {
	if (DEBUG)
	    printf("exiting URL base with NULL\n");
    }
    return base;
}

int isMms(char *url)
{
    if ((strncasecmp(url, "mms://", 6) == 0)
	|| (strncasecmp(url, "mmst://", 7) == 0)
	|| (strncasecmp(url, "rtsp://", 7) == 0)) {
	return 1;
    } else {
	return 0;
    }
}

void mmsToHttp(char *dest, char *src)
{
    char *temp;

    temp = (char *) NPN_MemAlloc(sizeof(char) * (strlen(src) + 1 +
						 sizeof("http") -
						 sizeof("mms")));
    strcpy(temp, "http");
    strcat(temp, src + 3);
    strcpy(dest, temp);
    NPN_MemFree(temp);
    return;
}

void sendCommand(FILE * control, char *command)
{
    if (control != NULL) {
	fprintf(control, "%s\n", command);
	fflush(control);
    }
}

int URLcmp(const char *url1, const char *url2)
{
    char *buffer1, *buffer2;
    char *tmp;

    if (strcmp(url1, url2) == 0)
	return 0;

    // replace %20 with spaces in both strings
    buffer1 = strdup(url1);
    buffer2 = strdup(url2);

    while ((tmp = strstr(buffer1, "%20"))) {
	if (tmp != NULL) {
	    *tmp = ' ';
	    tmp++;
	    *tmp = '\0';
	    strcat(buffer1, tmp + 2);
	}
    }

    while ((tmp = strstr(buffer2, "%20"))) {
	if (tmp != NULL) {
	    *tmp = ' ';
	    tmp++;
	    *tmp = '\0';
	    strcat(buffer2, tmp + 2);
	}
    }

    if (strcmp(buffer1, buffer2) == 0) {
	//if (DEBUG) printf("URL's match\n buffer1 %s len:%i\n buffer2 %s len %i\n",buffer1,strlen(buffer1),buffer2,strlen(buffer2));
	free(buffer1);
	free(buffer2);
	return 0;
    } else {
	//if (DEBUG) printf("URL's don't match\n buffer1 %s len:%i\n buffer2 %s len%i\n",buffer1,strlen(buffer1),buffer2,strlen(buffer2));
	free(buffer1);
	free(buffer2);
	return -1;
    }
}

// Remove all single quotes from url to prevent remote execution
void remove_quotes(char *url)
{
    char *p;

    // don't scan the string if it not in there, this should be a bit faster than the while loop

    p = strchr(url, '`');

    if (p == NULL) {
	return;
    } else {
	if (DEBUG) {
	    printf("single quotes in url (%s), truncating\n"
		   "WARNING: probably won't work! FIXME!\n", url);
	}
	*p = '\0';
    }
}
