#include "plugin.h"
#include <sys/types.h>
#include <sys/stat.h>

extern int DEBUG;
extern int DESTROYED;

FILE *mypopen(char **argv, pid_t * pid, FILE ** control)
{
    int filedesr[2], filedesw[2];
    pid_t child;

    pipe(filedesr);
    pipe(filedesw);
    child = fork();

    if (!child) {
	if (DEBUG) {
	    char **tmp;
	    tmp = argv;
	    printf("Starting: ");
	    while (*tmp)
		printf("%s ", *(tmp++));
	    printf("\n");
	}

	dup2(filedesw[0], 0);
	dup2(filedesr[1], 1);
	dup2(filedesr[1], 2);
	close(filedesw[1]);
	close(filedesr[0]);
	setsid();
	setpgid(0, 0);

	if (execvp(argv[0], argv) < 0)
	    perror("execv");
	_exit(0);
    } else {
	*pid = child;
	*control = fdopen(filedesw[1], "w");
	close(filedesw[0]);
	close(filedesr[1]);
	return fdopen(filedesr[0], "r");
    }
}


void SetupPlayer(nsPluginInstance * instance, XEvent * event)
{
    int i;
    char xval[32], yval[32];
    char *filename;
    char *baseurl;

    instance->setup = 1;
    instance->td->w = instance->widget;
    instance->td->instance = instance;

    DrawUI(instance->widget, instance, "Loading Media...", 0, -1);


    if (instance->href) {
	snprintf(instance->td->list->url, 1024, "%s", instance->href);
    } else if (instance->fname) {
	snprintf(instance->td->list->url, 1024, "%s", instance->fname);
    } else {
	snprintf(instance->td->list->url, 1024, "%s", instance->url);
    }

    if (instance->mode == NP_FULL) {
	snprintf(xval, 32, "%i", instance->window_width);
	snprintf(yval, 32, "%i", instance->window_height);
    } else {
	snprintf(xval, 32, "%i", instance->embed_width);
	snprintf(yval, 32, "%i", instance->embed_height);
    }

    baseurl = getURLBase(instance->td->list->url);
    if (baseurl != NULL) {
	if (instance->baseurl == NULL) {
	    instance->baseurl = baseurl;
	} else {
	    if (strcmp(instance->baseurl, baseurl) != 0) {
		NPN_MemFree(instance->baseurl);
		instance->baseurl = baseurl;
	    } else {
		NPN_MemFree(baseurl);
	    }
	}
    }

    if (instance->hostname == NULL)
	instance->hostname = getURLHostname(instance->td->list->url);

    if (instance->keep_download == 1) {
	instance->td->list->remove = 0;
	filename = getURLFilename(instance->td->list->url);
	snprintf(instance->td->list->fname, 1024, "%s/%s",
		 instance->download_dir, filename);
	if (filename)
	    NPN_MemFree(filename);
    } else {
	if (strlen(instance->td->list->fname) == 0) {
	    snprintf(instance->td->list->fname, 1024, "%s",
		     tempnam("/tmp", "mplayerplug-inXXXXXX"));
	}
    }

    i = 0;
    snprintf(instance->td->argv[i++], 1024, "mplayer");
    if (instance->novop == 1) {
	snprintf(instance->td->argv[i++], 1024, "-vop");
	snprintf(instance->td->argv[i++], 1024, "null");
    } else {
	if (instance->vop != NULL) {
	    snprintf(instance->td->argv[i++], 1024, "-vop");
	    snprintf(instance->td->argv[i++], 1024, "%s", instance->vop);
	    if (strncmp(instance->vop, "scale=", 6) == 0) {
		snprintf(instance->td->argv[i++], 1024, "-fs");
	    }
	}
    }


    if (instance->mode == NP_EMBED) {
	if (instance->window != 0) {
	    snprintf(instance->td->argv[i++], 1024, "-wid");
	    snprintf(instance->td->argv[i++], 1024, "0x%x",
		     (int) instance->window);
	}
    }


    if ((instance->embed_width == 0)
	&& (instance->mode == NP_EMBED)) {
	// do nothing
    } else {

	if (instance->maintain_aspect == 1) {
	    snprintf(instance->td->argv[i++], 1024, "-xy");
	    snprintf(instance->td->argv[i++], 1024, "%s", xval);
	} else {
	    snprintf(instance->td->argv[i++], 1024, "-x");
	    snprintf(instance->td->argv[i++], 1024, "%s", xval);
	    snprintf(instance->td->argv[i++], 1024, "-y");
	    snprintf(instance->td->argv[i++], 1024, "%s", yval);
	}

    }

    if (instance->rtsp_use_tcp == 1) {
	snprintf(instance->td->argv[i++], 1024, "-rtsp-stream-over-tcp");
    }

    if (instance->cachesize > 0) {
	snprintf(instance->td->argv[i++], 1024, "-cache");
	snprintf(instance->td->argv[i++], 1024, "%i", instance->cachesize);
    } else {
	snprintf(instance->td->argv[i++], 1024, "-nocache");
    }

    if (instance->loop == 1) {
	snprintf(instance->td->argv[i++], 1024, "-loop");
	snprintf(instance->td->argv[i++], 1024, "0");
    }

    if (instance->vo) {
	snprintf(instance->td->argv[i++], 1024, "-vo");
	snprintf(instance->td->argv[i], 1024, "%s", instance->vo);
	if ((strcmp(instance->td->argv[i], "x11") == 0)
	    || (strstr(instance->td->argv[i], ",x11") != NULL)) {
	    i++;
	    snprintf(instance->td->argv[i++], 1024, "-zoom");	/* -vo x11 needs -zoom for rescaling */
	} else {
	    i++;
	}
    }

    if (instance->ao) {
	snprintf(instance->td->argv[i++], 1024, "-ao");
	snprintf(instance->td->argv[i++], 1024, "%s", instance->ao);
    }

    if (instance->output_display) {
	snprintf(instance->td->argv[i++], 1024, "-display");
	snprintf(instance->td->argv[i++], 1024, "%s",
		 instance->output_display);
    }

    snprintf(instance->td->argv[i++], 1024, "-osdlevel");
    snprintf(instance->td->argv[i++], 1024, "%i", instance->osdlevel);

    /* some extra flags for mplayer */
    snprintf(instance->td->argv[i++], 1024, "-nojoystick");	/* annoying... */
    snprintf(instance->td->argv[i++], 1024, "-nofs");	/* no full screen */
    snprintf(instance->td->argv[i++], 1024, "-slave");	/* no full screen */

    /* I had to remove this option since I need to know the currently 
       played second (pipas)
       
    snprintf(instance->td->argv[i++], 1024, "-quiet");*/

    memset(instance->td->argv[i++], '\0', 1);


    // ok we have the command line setup
    if (DEBUG)
	printf("ready to setup threads\n");

    pthread_attr_setdetachstate(&(instance->thread_attr),
				PTHREAD_CREATE_DETACHED);

    if (DEBUG)
	printf("creating thread - NP_EMBED\n");
    DrawUI(instance->widget, instance, "Getting playlist...", 0, -1);
    if (instance->state < STATE_GETTING_PLAYLIST)
	instance->state = STATE_GETTING_PLAYLIST;
    if (DEBUG)
	printf("creating player thread\n");
    pthread_create(&(instance->player_thread), &(instance->thread_attr),
		   playPlaylist, (void *) instance->td);
    // recommended slight pause
    usleep(1);

    if (DEBUG)
	printf("MAIN THREAD DONE\n");
}

#define SMILDEBUG

void refresh_frame( char *buffer, _ThreadData *td, Node *node )
{
	int seconds;
	static int oldseconds = -1;
	char *start;
	char *endptr;
	area *runner;
#ifdef SMILDEBUG
	static FILE *error = NULL;
	char errortmp[64];
	if (!error)
		error = fopen("/tmp/mplayerplugin-smil.log", "a");

        fprintf( error,
		"url: %s\nfname: %s\n status: %d\nretrieved: %d\n"
		"playlist: %d\n bytes: %ld\ntotalbytes:%ld\n"
		"cachebytes: %ld",
		node->url, node->fname, node->status, node->retrieved, 
		node->playlist, node->bytes, node->totalbytes );
#endif

	/* just to be sure */
	if( node == NULL || node->area == NULL )
		return;

	start = buffer;	
	while( (start = strstr(start, "A:")) != NULL && strlen(start)>7 ){
		start += 2;
		seconds = (int)strtol( start, &endptr, 0 );
		if (seconds == oldseconds || start == endptr)
			continue;
#ifdef SMILDEBUG
		strncpy(errortmp, start, 63);
		errortmp[64] = 0;
		fprintf(error, "->seconds: %d, oldseconds: %d, buffer: '%s'\n",
				seconds, oldseconds, errortmp );
		fflush(error);
#endif
		runner = node->area;
		while( runner ){
			if(runner->begin == seconds){
				NPN_GetURL( td->instance->mInstance, 
						runner->url, 
						runner->target );
				break;
			}
			runner = runner->next;
		}
		oldseconds = seconds;
	}
}


void *playPlaylist(void *td)
{

    //FILE *player;
    FILE *playlist;
    char buffer[1024];
    char cmd[2048];
    char message[1024];
    char mmsplaylist[1024];
    char *eos;
    char *argv[256];
    int argc = 0;
    ThreadData *local_td;
    Node *local_list, *copy;
    int i;
    Node *node;
    int notfound;
    char *vo;
    char vm[10];
    char *cf;
    int cfpercent;
    FILE *error = fopen( "/tmp/mplayerplugin-smil.log", "a");

    local_td = (ThreadData *) td;
    local_list = local_td->list;


    if (DEBUG)
	printf("playPlayList - deciding to wait\n");

    local_td->instance->js_state = JS_STATE_BUFFERING;
    pthread_mutex_lock(&(local_td->instance->playlist_mutex));
    if (local_td->instance->state < STATE_STARTED_PLAYER) {
	pthread_cond_wait(&(local_td->instance->playlist_complete_cond),
			  &(local_td->instance->playlist_mutex));
	local_td->instance->state = STATE_STARTED_PLAYER;
    }
	
		

    
    pthread_mutex_unlock(&(local_td->instance->playlist_mutex));

    if (DEBUG)
	printf("playPlayList - waiting complete\n");


    DrawUI(local_td->w, local_td->instance, "Loading Media...", 0, -1);

    memset(cmd, '\0', 1);
    i = 0;
    if (DEBUG)
	printf("building command string\n");

    while (strcmp(local_td->argv[i], "") != 0) {
	argv[argc++] = local_td->argv[i];
	if (DEBUG)
	    printf("PLAY %i:%s\n", i, local_td->argv[i]);
	i++;
    }

    i = 0;

    while (local_td->instance->qtNext[i] != NULL) {
	node = newNode();
	if (DEBUG)
	    printf("adding %s\n", local_td->instance->qtNext[i]);
	strlcpy(node->url, local_td->instance->qtNext[i],
		sizeof(node->url));
	i++;
	addToEnd(local_list, node);
    }

    // preview playlist
    if (DEBUG) {
	while (local_list != NULL) {
	    if (DEBUG)
		printf("PL URL: %s\n", local_list->url);
	    local_list = local_list->next;
	}
	local_list = local_td->list;
    }

    copy = newNode();
    pthread_mutex_lock(&(local_td->instance->playlist_mutex));
    copyNode(copy, local_list);
    pthread_mutex_unlock(&(local_td->instance->playlist_mutex));
    
	/*
    	while( copy->retrieved==0 ){
		fprintf( error, "not retrieved!\n");
    		pthread_mutex_lock(&(local_td->instance->playlist_mutex));
		pthread_cond_wait(&(local_td->instance->playlist_complete_cond),
			  &(local_td->instance->playlist_mutex));
	}
	*/

    while (copy != NULL) {
	if (DEBUG)
	    printf("url: %s \n play %i\n", copy->url, copy->play);
	strlcpy(message, "Playing ", sizeof(message));
	strlcat(message, copy->url, sizeof(message));
	DrawUI(local_td->instance->widget, local_td->instance, message, 0,
	       -1);
	if (copy->play) {

	    // for rtsp streams we need to specify FPS
	    if (strncmp(copy->url, "rtsp", 4) == 0) {
		argv[argc++] = "-fps";
		argv[argc++] = "30";
	    }
	    if (copy->playlist == 1)
		argv[argc++] = "-playlist";

	    if (copy->mmsstream) {
		argv[argc++] = copy->url;
		if (local_td->instance->keep_download) {
		    snprintf(mmsplaylist, sizeof(mmsplaylist),
			     "%s/playlist",
			     local_td->instance->download_dir);
		    playlist = fopen(mmsplaylist, "a");
		    if (playlist != NULL) {
			fprintf(playlist, "%s\n", copy->url);
			fclose(playlist);
		    }
		}
	    } else {
		if (strlen(copy->fname) == 0) {
		    argv[argc++] = copy->url;
		} else {
		    argv[argc++] = copy->fname;
		}
	    }

	    if (DEBUG) {
		printf("URL: %s\n", copy->url);
	    }

	    if (!copy->mmsstream) {
		if ((copy->bytes < (local_td->instance->cachesize * 1024))
		    && (copy->retrieved == 0)) {
		    // perhaps we need to wait here in case the cache is not full
		}
	    }
	    
	    if( local_list->area )
	    while( local_list->retrieved==0 ){
    	fprintf( error, "url:%s\n"
			"retrieved?: %d\nbytes:      %ld\ntotalbytes: %ld\n", 
			local_list->url, local_list->retrieved, local_list->bytes,
			local_list->totalbytes);
	fflush( error );
	sleep(1);
	    }


	    argv[argc++] = NULL;
	    local_td->instance->player =
		mypopen(argv, &local_td->instance->pid,
			&local_td->instance->control);

	    local_td->instance->state = STATE_PLAYING;
	    if (local_td->instance->player != NULL) {
		local_td->instance->js_state = JS_STATE_PLAYING;
		while (fgets
		       (buffer, sizeof(buffer),
			local_td->instance->player) != NULL) {
		    if (DEBUG) {
			printf("READ: %s \n", buffer);
		    }

		refresh_frame( buffer, local_td, local_list );

		    if (strstr(buffer, "Cache fill:") != NULL) {
			strlcpy(message, strstr(buffer, "Cache fill"),
				sizeof(message));
			notfound = 1;
			while (notfound) {
			    eos = strrchr(message, '\r');
			    if (eos == NULL)
				eos = message;
			    if (strstr(eos, "Cache fill:") == NULL) {
				// if we don't find "Cache fill:" then replace the \r with \0, making message shorter
				eos[0] = '\0';
				// break out of the loop if too short
				if (strlen(message) < 5)
				    break;
			    } else {
				// if we found the last one in the message string, move from \r to C in the string
				eos++;
				notfound = 0;
			    }
			}
			// Only update the display if there is something worth displaying
			if (strstr(eos, "Cache fill:") != NULL) {
			    cf = strstr(eos, "Cache fill:");
			    sscanf(cf, "Cache fill: %i", &cfpercent);
			    strlcpy(message, "Caching ", sizeof(message));
			    strlcat(message, copy->url, sizeof(message));
			    DrawUI(local_td->instance->widget,
				   local_td->instance, eos, -1, cfpercent);
			}
		    }

		    if (strstr(buffer, "Starting") != NULL) {
			strlcpy(message, "Playing ", sizeof(message));
			strlcat(message, copy->url, sizeof(message));
			DrawUI(local_td->instance->widget,
			       local_td->instance, message, 0, -1);
		    }

		    if (strstr(buffer, "VO:") != NULL) {
			if (local_td->instance->mode != NP_FULL)
			    local_td->instance->noredraw = 1;
			vo = strstr(buffer, "VO:");
			sscanf(vo, "VO: [%10[^]]] %ix%i => %ix%i", vm,
			       &(local_list->actual_x),
			       &(local_list->actual_y),
			       &(local_list->play_x),
			       &(local_list->play_y));
		    }

		    if (strstr(buffer, "Error while decoding") != NULL) {
			if (DEBUG)
			    printf("Resetting stream, 1 sec rewind\n");
			sleep(1);
			sendCommand(local_td->instance->control,
				    "seek -1 0\n");
		    }

		    if (strstr(buffer, "Quit") != NULL) {
			if (DEBUG)
			    printf("breaking read loop - Quit\n");
			break;
		    }

		    if (strstr(buffer, "Exiting") != NULL) {
			if (DEBUG)
			    printf("breaking read loop - Exiting\n");
			break;
		    }
		    // break out if mplayer crashes
		    if (strstr(buffer, "interrupted") != NULL) {
			if (DEBUG)
			    printf("breaking read loop - interrupted\n");
			break;
		    }

		    if (strstr(buffer, "explicit kill") != NULL) {
			if (DEBUG)
			    printf("breaking read loop - killed\n");
			break;
		    }

		    if (local_td->instance->player == NULL) {
			if (DEBUG)
			    printf
				("breaking read loop - player == NULL\n");
			break;
		    }

		    if (local_td->instance->cancelled == 1) {
			if (DEBUG)
			    printf("breaking read loop - cancelled\n");
			break;
		    }
		}
		local_list->cancelled = 1;
		if (local_td->instance->control != NULL) {
		    fclose(local_td->instance->control);
		    local_td->instance->control = NULL;
		}
		if (local_td->instance->player != NULL) {
		    fclose(local_td->instance->player);
		    local_td->instance->player = NULL;
		}

		if (DEBUG)
		    printf("close done\n");
		local_td->instance->state = STATE_PLAYLIST_NEXT;
	    }
	}
	local_td->instance->js_state = JS_STATE_TRANSITIONING;
	local_td->instance->noredraw = 0;
	if (DESTROYED) {
	    local_td->instance->js_state = JS_STATE_UNDEFINED;
	    return NULL;
	}
	if (local_td->instance != NULL) {
	    if (DESTROYED) {
		local_td->instance->js_state = JS_STATE_UNDEFINED;
		return NULL;
	    }
	    if (local_td->instance->state != STATE_PLAY_CANCELLED) {
		if (DESTROYED) {
		    local_td->instance->js_state = JS_STATE_UNDEFINED;
		    return NULL;
		}
		if (local_list != NULL) {
		    if (DESTROYED) {
			local_td->instance->js_state = JS_STATE_UNDEFINED;
			return NULL;
		    }
		    local_list = local_list->next;
		    if (local_list == NULL) {
			deleteNode(copy);
			copy = NULL;
			local_td->instance->js_state = JS_STATE_STOPPED;
		    } else {
			copyNode(copy, local_list);
		    }
		}
	    } else {
		break;
		copy = NULL;
	    }
	}
    }
    local_td->instance->js_state = JS_STATE_UNDEFINED;
    return NULL;
}
