#include "plugin.h"
#include <errno.h>

extern int errno;
extern int DEBUG;
extern int DESTROYED;
#include <X11/xpm.h>
#include <X11/cursorfont.h>
#include "../pixmaps/logo.xpm"
#include "../pixmaps/progress_left.xpm"
#include "../pixmaps/progress_middle.xpm"
#include "../pixmaps/progress_right.xpm"
#include "../pixmaps/progress_fill.xpm"

#define BORDER               10
#define LOGO_WIDTH          348
#define LOGO_HEIGHT          48

#ifndef HAVE_X11_XPM
#error libXpm has not been found. Compilation cannot continue
#endif

static XFontStruct *font = NULL;
static Pixmap logo, logomask;
static Pixmap progress_left = (Pixmap) NULL, progress_leftmask =
    (Pixmap) NULL;
static Pixmap progress_middle = (Pixmap) NULL, progress_middlemask =
    (Pixmap) NULL;
static Pixmap progress_right = (Pixmap) NULL, progress_rightmask =
    (Pixmap) NULL;
static Pixmap progress_fill = (Pixmap) NULL, progress_fillmask =
    (Pixmap) NULL;
static int lastpercent = -1;

void Initialize(Display * dpy, Drawable d, GC gc)
{
    Cursor guicursor;
    XSetWindowAttributes attrs;

    XpmCreatePixmapFromData(dpy, d, logo_xpm, &logo, &logomask, NULL);
    XpmCreatePixmapFromData(dpy, d, progress_left_xpm, &progress_left,
			    &progress_leftmask, NULL);
    XpmCreatePixmapFromData(dpy, d, progress_middle_xpm, &progress_middle,
			    &progress_middlemask, NULL);
    XpmCreatePixmapFromData(dpy, d, progress_right_xpm, &progress_right,
			    &progress_rightmask, NULL);
    XpmCreatePixmapFromData(dpy, d, progress_fill_xpm, &progress_fill,
			    &progress_fillmask, NULL);
    font = XQueryFont(dpy, XGContextFromGC(gc));
    guicursor = XCreateFontCursor(dpy, XC_watch);
    attrs.cursor = guicursor;
    XChangeWindowAttributes(dpy, d, CWCursor, &attrs);
}

void DrawLogo(Display * dpy, Drawable d, GC gc)
{
    XGCValues values;

    values.clip_mask = logomask;
    values.clip_x_origin = 0;
    values.clip_y_origin = 0;
    XChangeGC(dpy, gc, GCClipMask | GCClipXOrigin | GCClipYOrigin,
	      &values);
    XCopyArea(dpy, logo, d, gc, 0, 0, LOGO_WIDTH, LOGO_HEIGHT, 0, 0);
    values.clip_mask = (Pixmap) NULL;
    XChangeGC(dpy, gc, GCClipMask | GCClipXOrigin | GCClipYOrigin,
	      &values);
}


void DrawProgressBG(Display * dpy, Window w, int x, int y, int width,
		    int height)
{
    XGCValues values;
    GC border_gc;

    border_gc = XCreateGC(dpy, w, GCForeground, &values);

    values.clip_mask = progress_leftmask;
    values.clip_x_origin = x - 10;
    values.clip_y_origin = y - 10;
    XChangeGC(dpy, border_gc, GCClipMask | GCClipXOrigin | GCClipYOrigin,
	      &values);
    XCopyArea(dpy, progress_left, (Drawable) w, border_gc, 0, 0, 10, 31,
	      x - 10, y - 10);
    values.clip_mask = (Pixmap) NULL;
    XChangeGC(dpy, border_gc, GCClipMask | GCClipXOrigin | GCClipYOrigin,
	      &values);

    XSetTile(dpy, border_gc, progress_middle);
    XSetFillStyle(dpy, border_gc, FillTiled);
    XSetTSOrigin(dpy, border_gc, x, y - 10);
    XFillRectangle(dpy, w, border_gc, x, y - 10, width, 31);
    XSetFillStyle(dpy, border_gc, FillSolid);

    values.clip_mask = progress_rightmask;
    values.clip_x_origin = x + width;
    values.clip_y_origin = y - 10;
    XChangeGC(dpy, border_gc, GCClipMask | GCClipXOrigin | GCClipYOrigin,
	      &values);
    XCopyArea(dpy, progress_right, (Drawable) w, border_gc, 0, 0, 10, 31,
	      x + width, y - 10);
    values.clip_mask = (Pixmap) NULL;
    XChangeGC(dpy, border_gc, GCClipMask | GCClipXOrigin | GCClipYOrigin,
	      &values);

    XFreeGC(dpy, border_gc);
}

void DrawProgress(Display * dpy, Window w, int x, int y, int width,
		  int height, int percent)
{
    GC progress_gc;
    int pixelpercent;
    XGCValues values;

    if (percent > 100)
	percent = 100;
    progress_gc = XCreateGC(dpy, w, GCForeground, &values);
    XSetTile(dpy, progress_gc, progress_fill);
    XSetFillStyle(dpy, progress_gc, FillTiled);
    XSetTSOrigin(dpy, progress_gc, x, y);
    pixelpercent = (int) ((width - x) * (percent / 100.));
    XFillRectangle(dpy, w, progress_gc, x, y, pixelpercent, height);
    XSetFillStyle(dpy, progress_gc, FillSolid);
    XFreeGC(dpy, progress_gc);
}



void DrawUI(Widget w, nsPluginInstance * instance, char *message,
	    int FullRedraw, int percent)
{

    GC black_gc, white_gc;
    XGCValues values;
    const char *id = "mplayerplug-in v" PACKAGE_VERSION;
    unsigned int id_width;
    int top, font_height, loops;
    int term, chop;
    int PercentRedraw;

    if (0) {
	printf("Widget: %i\n", w);
	printf("Message: %s\n", message);
    }

    if (w == NULL)
	return;
    if (instance->state < STATE_READY)
	return;
    if (message == NULL)
	return;
    if (DESTROYED == 1)
	return;
    if (instance->display == NULL)
	return;

    sprintf(instance->lastmessage, "%s", message);

    values.foreground =
	BlackPixel((Display *) instance->display,
		   DefaultScreen((Display *) instance->display));
    black_gc =
	XCreateGC((Display *) instance->display,
		  (Drawable) instance->window, GCForeground, &values);
    values.foreground =
	WhitePixel((Display *) instance->display,
		   DefaultScreen((Display *) instance->display));
    white_gc =
	XCreateGC((Display *) instance->display,
		  (Drawable) instance->window, GCForeground, &values);

    PercentRedraw = FALSE;
    if (percent != -1 && lastpercent != -1 && lastpercent > percent)
	PercentRedraw = TRUE;
    if (font == NULL) {
	FullRedraw = TRUE;
	Initialize((Display *) instance->display,
		   (Drawable) instance->window, black_gc);
    }
    font_height = font->max_bounds.ascent + font->max_bounds.descent;
    id_width = XTextWidth(font, id, strlen(id));
    if (FullRedraw || instance->window_height <= 80) {
	XFillRectangle((Display *) instance->display,
		       (Drawable) instance->window, white_gc, 0, 0,
		       instance->window_width, instance->window_height);
    } else {
	XFillRectangle((Display *) instance->display,
		       (Drawable) instance->window, white_gc, BORDER,
		       LOGO_HEIGHT, instance->window_width - BORDER * 2,
		       instance->window_height - font_height -
		       LOGO_HEIGHT - BORDER - 10);
    }
    if (instance->window_height > 80
	&& instance->window_width > id_width + BORDER * 2) {
	top = BORDER * 2 + LOGO_HEIGHT;
	loops = 0;
	if (FullRedraw) {
	    DrawLogo((Display *) instance->display,
		     (Drawable) instance->window, black_gc);
	    XDrawString((Display *) instance->display,
			(Drawable) instance->window, black_gc,
			instance->window_width - id_width - BORDER,
			instance->window_height - font_height, id,
			strlen(id));
	}
	if ((FullRedraw || PercentRedraw)
	    && instance->window_width > id_width + BORDER * 3 + 100)
	    // The height is set to 11 (my value for font_height)
	    // since the progress image only has 11 pixels of space
	    DrawProgressBG((Display *) instance->display,
			   (Drawable) instance->window, BORDER,
			   instance->window_height - font_height - BORDER,
			   instance->window_width - id_width - BORDER * 3,
			   11);
	// Loop a max of 10 times in case of problems (like too small of a width)
	while (strlen(message) != 0 && loops < 10) {
	    chop = 0;
	    term = 0;
	    while (XTextWidth(font, message, strlen(message) - chop) +
		   BORDER * 2 > (int) instance->window_width)
		chop++;
	    if (strchr(message, '\n') != NULL &&
		(int) (strchr(message, '\n') - message) <
		(int) (strlen(message) - chop)) {
		chop = strlen(message) - (strchr(message, '\n') - message);
		term = 1;
	    }
	    XDrawString((Display *) instance->display,
			(Drawable) instance->window, black_gc, BORDER, top,
			message, strlen(message) - chop);
	    message += strlen(message) - chop + term;
	    top += font_height;
	    loops++;
	    if (top > (int) instance->window_height - font_height - BORDER)
		break;
	}
	// Make sure we have at least 100 px in addition to the id tag and borders
	if (percent != -1
	    && instance->window_width > id_width + BORDER * 3 + 100)
	    DrawProgress((Display *) instance->display,
			 (Drawable) instance->window, BORDER,
			 instance->window_height - font_height - BORDER,
			 instance->window_width - id_width - BORDER * 3,
			 11, percent);
	else if (FullRedraw && lastpercent != -1
		 && instance->window_width > id_width + BORDER * 3 + 100)
	    DrawProgress((Display *) instance->display,
			 (Drawable) instance->window, BORDER,
			 instance->window_height - font_height - BORDER,
			 instance->window_width - id_width - BORDER * 3,
			 11, lastpercent);

    } else {
	XDrawString((Display *) instance->display,
		    (Drawable) instance->window, black_gc, BORDER, 15,
		    message, strlen(message));
    }
    if (percent != -1)
	lastpercent = percent;

    XFreeGC((Display *) instance->display, black_gc);
    XFreeGC((Display *) instance->display, white_gc);

    XFlush((Display *) instance->display);
    usleep(100);
}

void FreeUI(Display * dpy)
{
    font = NULL;
    // Free all the pixmaps used by the GUI
    if (progress_left)
	XFreePixmap(dpy, progress_left);
    if (progress_leftmask)
	XFreePixmap(dpy, progress_leftmask);
    if (progress_middle)
	XFreePixmap(dpy, progress_middle);
    if (progress_middlemask)
	XFreePixmap(dpy, progress_middlemask);
    if (progress_right)
	XFreePixmap(dpy, progress_right);
    if (progress_rightmask)
	XFreePixmap(dpy, progress_rightmask);
    if (progress_fill)
	XFreePixmap(dpy, progress_fill);
    if (progress_fillmask)
	XFreePixmap(dpy, progress_fillmask);
    // Don't free the font since we didn't create it
}

void DestroyCB(Widget widget, XtPointer client_data, XtPointer call_data)
{
    nsPluginInstance *instance = (nsPluginInstance *) client_data;
    int count, status;

    if (DEBUG)
	printf("window destroy callback\n");

    if (instance == NULL)
	return;

    if (widget == instance->widget) {
	DESTROYED = 1;

	if (instance->pid != 0) {
	    sendCommand(instance->control, "quit\n");
	    instance->cancelled = 1;
	}
	count = 0;
	while ((instance->player != NULL) && (count < 10)) {
	    usleep(100);
	    count++;
	    if (DEBUG)
		printf("waiting for player to go NULL %i\n", count);
	}
	if (instance->player == NULL)
	    instance->pid = 0;

	pthread_mutex_lock(&(instance->playlist_mutex));
	// Free any UI data necessary
	FreeUI((Display *) instance->display);
	XtRemoveCallback(instance->widget, XtNdestroyCallback,
			 DestroyCB, (XtPointer) instance);
	XtRemoveEventHandler(instance->widget, ExposureMask, FALSE,
			     (XtEventHandler) RedrawCB, instance);
	instance->widget = NULL;
	instance->state = STATE_PLAY_CANCELLED;

	if (DEBUG)
	    printf("starting wait loop\n");

	if (instance->pid != 0) {
	    if (DEBUG)
		printf
		    ("Trying to kill mplayer process(%d), if it still exists\n",
		     instance->pid);

	    count = 0;
	    status = 1;
	    while ((status != 0) && (count < 10)) {
		status = kill(instance->pid, 15);
		if (DEBUG)
		    printf("kill(15) status = %i\n", status);
		if (errno == ESRCH) {
		    status = 0;
		    if (DEBUG)
			printf("PID was not found breaking out.\n",
			       status);
		    break;
		}
		usleep(100);
		count++;
	    }

	    if (status != 0) {
		status = kill(instance->pid, 9);
		if (status == 0) {
		    instance->pid = 0;
		}
		if (DEBUG)
		    printf("kill(9) status = %i\n", status);
	    }

	    if (instance->control) {
		fclose(instance->control);
		instance->control = NULL;
		if (DEBUG)
		    printf("control closed\n");
	    }

	}

	deleteList(instance->td->list);

	if (instance->td != NULL) {
	    NPN_MemFree(instance->td);
	    instance->td = NULL;
	}
	if (instance->baseurl != NULL) {
	    NPN_MemFree(instance->baseurl);
	    instance->baseurl = NULL;
	}
	if (instance->hostname != NULL) {
	    NPN_MemFree(instance->hostname);
	    instance->hostname = NULL;
	}


	instance->td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
	instance->td->list = newNode();
	instance->td->instance = NULL;
	instance->lastmessage = (char *) NPN_MemAlloc(sizeof(char) * 1024);
	memset(instance->lastmessage, '\0', 1);
	pthread_mutex_unlock(&(instance->playlist_mutex));
    } else {
	if (DEBUG)
	    printf("Widgets don't match %i != %i\n", widget,
		   instance->widget);
    }

}

void RedrawCB(Widget widget, XtPointer client_data, XtPointer call_data)
{
    char message[1024];

    nsPluginInstance *instance = (nsPluginInstance *) client_data;
    if (DEBUG)
	printf("redrawing window\n");

    if (DESTROYED || instance->noredraw)
	return;

    if (instance != NULL) {
	if (instance->lastmessage != NULL) {
	    if (strlen(instance->lastmessage) != 0) {
		if (instance->state >= STATE_GETTING_PLAYLIST) {
		    strcpy(message, instance->lastmessage);
		    DrawUI(instance->widget, instance, message, 1, -1);
		}
	    }
	}
    }
}
