/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "plugin.h"
#include "nsIServiceManager.h"
#include "nsIMemory.h"
#include "nsISupportsUtils.h"	// this is where some useful macros defined
#include <errno.h>

extern int errno;
extern int DEBUG;
extern int DESTROYED;

int32 STREAMBUFSIZE = 0X0FFFFFFF;

// service manager which will give the access to all public browser services
// we will use memory service as an illustration
nsIServiceManager *gServiceManager = NULL;

// Unix needs this
#ifdef XP_UNIX

char *NPP_GetMIMEDescription(void)
{
    return GetMIMEDescription();
}

// get values per plugin
NPError NS_PluginGetValue(NPPVariable aVariable, void *aValue)
{
    return GetValue(aVariable, aValue);
}
#endif				//XP_UNIX

//////////////////////////////////////
//
// general initialization and shutdown
//
NPError NS_PluginInitialize()
{
    // this is probably a good place to get the service manager
    // note that Mozilla will add reference, so do not forget to release
    nsISupports *sm = NULL;


    NPN_GetValue(NULL, NPNVserviceManager, &sm);

    // Mozilla returns nsIServiceManager so we can use it directly; doing QI on
    // nsISupports here can still be more appropriate in case something is changed
    // in the future so we don't need to do casting of any sort.
    if (sm) {
	sm->QueryInterface(NS_GET_IID(nsIServiceManager),
			   (void **) &gServiceManager);
	NS_RELEASE(sm);
    }

    return NPERR_NO_ERROR;
}

void NS_PluginShutdown()
{
    // we should release the service manager
    NS_IF_RELEASE(gServiceManager);
    gServiceManager = NULL;
}

/////////////////////////////////////////////////////////////
//
// construction and destruction of our plugin instance object
//
nsPluginInstanceBase *NS_NewPluginInstance(nsPluginCreateData *
					   aCreateDataStruct)
{
    if (!aCreateDataStruct)
	return NULL;

    nsPluginInstance *plugin =
	new nsPluginInstance(aCreateDataStruct->instance);

    New(plugin, aCreateDataStruct);
    return plugin;
}

void NS_DestroyPluginInstance(nsPluginInstanceBase * aPlugin)
{
    if (aPlugin)
	delete(nsPluginInstance *) aPlugin;
}

////////////////////////////////////////
//
// nsPluginInstance class implementation
//
nsPluginInstance::nsPluginInstance(NPP aInstance):nsPluginInstanceBase(),
mInstance(aInstance),
mInitialized(FALSE), mScriptablePeer(NULL)
{
    int i;

    mScriptablePeer = getScriptablePeer();

    // init the instance variables here.
    mimetype = NULL;
    state = 0;
    url = NULL;
    fname = NULL;
    href = NULL;
    lastmessage = (char *) NPN_MemAlloc(sizeof(char) * 1024);
    memset(lastmessage, '\0', 1);
    mode = 0;
    window_width = 0;
    window_height = 0;
    embed_width = 0;
    embed_height = 0;
    setwindow = 0;
    baseurl = NULL;
    hostname = NULL;
    pid = 0;
    noredraw = 0;
    setup = 0;
    hrefrequested = 0;
    threadsignaled = 0;
    cancelled = 0;
    list = newNode();
    td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
    td->list = newNode();
    td->instance = NULL;
    control = NULL;
    player = NULL;
    DESTROYED = 0;
    autostart = 1;
    js_state = JS_STATE_UNDEFINED;
    nQtNext = 0;
    for (i = 0; i < 256; i++)
	qtNext[i] = NULL;
    widget = 0;
    display = NULL;
    window = 0;
    controlwindow = 0;

    // options
    vo = NULL;
    vop = NULL;
    novop = 0;
    ao = NULL;
    loop = 0;
    rtsp_use_tcp = 0;
    keep_download = 0;
    maintain_aspect = 1;
    download_dir = strdup(getenv("HOME"));
    cachesize = 512;
    output_display = NULL;
    osdlevel = 0;
    qt_speed = SPEED_MED;
    cache_percent = 25;

    // JavaScript
    paused = 0;

    // thread setup
    pthread_mutex_init(&playlist_mutex, NULL);
    pthread_attr_init(&thread_attr);
    pthread_cond_init(&playlist_complete_cond, NULL);


    LoadConfigFile(this);

}

nsPluginInstance::~nsPluginInstance()
{
    // mScriptablePeer may be also held by the browser
    // so releasing it here does not guarantee that it is over
    // we should take precaution in case it will be called later
    // and zero its mPlugin member
    if (DEBUG)
	printf("~nsPluginInstance called\n");

    // cleanup streams and files
    pthread_mutex_unlock(&playlist_mutex);
    pthread_attr_destroy(&thread_attr);
    pthread_mutex_destroy(&playlist_mutex);
    pthread_cond_destroy(&playlist_complete_cond);

    if (download_dir != NULL) {
	free(download_dir);
	download_dir = NULL;
    }

    if (td != NULL) {
	NPN_MemFree(td);
	td = NULL;
    }

    if (lastmessage != NULL) {
	NPN_MemFree(lastmessage);
	lastmessage = NULL;
    }

    if (mScriptablePeer != NULL)
	mScriptablePeer->SetInstance(NULL);
    NS_IF_RELEASE(mScriptablePeer);
}

NPBool nsPluginInstance::init(NPWindow * aWindow)
{
    if (aWindow == NULL)
	return FALSE;

    mInitialized = TRUE;
    return TRUE;
}

void nsPluginInstance::shut()
{

    int count, status, i;

    if (DEBUG)
	printf("shut called\n");

    pthread_mutex_lock(&playlist_mutex);
    DESTROYED = 1;

    if ((window_height == 0) && (window_width == 0) && (pid != 0)) {

	if (DEBUG)
	    printf
		("Trying to kill mplayer process(%d), if it still exists\n",
		 pid);

	sendCommand(control, "quit\n");
	cancelled = 1;
	count = 0;
	while (player != NULL && count < 10) {
	    usleep(100);
	    count++;
	}
	if (player == NULL)
	    pid = 0;

	if (pid != 0) {
	    count = 0;
	    status = 1;
	    while ((status != 0) && (count < 10)) {
		status = kill(pid, 15);
		if (DEBUG)
		    printf("kill(15) status = %i\n", status);
		if (errno == ESRCH) {
		    status = 0;
		    break;
		}

		usleep(100);
		count++;
	    }

	    if (status != 0) {
		status = kill(pid, 9);
		if (DEBUG)
		    printf("kill(9) status = %i\n", status);
		if (status == 0) {
		    pid = 0;
		}
	    }

	    if (control) {
		fclose(control);
		control = NULL;
	    }

	}
    }
    if (DEBUG)
	printf("mplayer dead\n");

    if (mimetype != NULL) {
	free(mimetype);
	mimetype = NULL;
    }

    if (url != NULL) {
	free(url);
	url = NULL;
    }

    if (fname != NULL) {
	free(fname);
	fname = NULL;
    }

    if (url != NULL) {
	free(url);
	url = NULL;
    }

    if (baseurl != NULL) {
	NPN_MemFree(baseurl);
	baseurl = NULL;
    }

    if (hostname != NULL) {
	NPN_MemFree(hostname);
	hostname = NULL;
    }

    if (vo != NULL) {
	free(vo);
	vo = NULL;
    }

    if (vop != NULL) {
	free(vop);
	vop = NULL;
    }

    if (ao != NULL) {
	free(ao);
	ao = NULL;
    }

    if (output_display != NULL) {
	free(output_display);
	output_display = NULL;
    }

    nQtNext = 0;
    for (i = 0; i < 256; i++) {
	if (qtNext[i] != NULL)
	    free(qtNext[i]);
	qtNext[i] = NULL;
    }

    if (DEBUG)
	printf("memory free\n");

    DESTROYED = 0;
    autostart = 1;
    cancelled = 0;
    js_state = JS_STATE_UNDEFINED;
    mInitialized = FALSE;
}

NPBool nsPluginInstance::isInitialized()
{
    return mInitialized;
}

NPError nsPluginInstance::SetWindow(NPWindow * aWindow)
{
    NPSetWindowCallbackStruct *ws;

    if ((aWindow == NULL) || (aWindow->window == NULL)) {
	return NPERR_NO_ERROR;
    }


    if ((Window) window != (Window) aWindow->window) {
	if (DEBUG)
	    printf("New window! old: 0x%x    new 0x%x\n",
		   window, aWindow->window);
    }

    if (controlwindow == 1)
	return NPERR_NO_ERROR;

    if (state < STATE_WINDOWSET) {
	if (DEBUG) {
	    printf("Size: %d %d %x\n", aWindow->x, aWindow->y,
		   aWindow->window);
	    printf("Size: %dx%d \n", aWindow->width, aWindow->height);
	}

	ws = (NPSetWindowCallbackStruct *) aWindow->ws_info;

	widget =
	    XtWindowToWidget((Display *) ws->display,
			     (Window) aWindow->window);

	XtAddCallback(widget, XtNdestroyCallback, DestroyCB,
		      (XtPointer) this);
	XtAddEventHandler(widget, ExposureMask, FALSE,
			  (XtEventHandler) RedrawCB, this);

	display = ws->display;
	window = (Window) aWindow->window;
	window_width = aWindow->width;
	window_height = aWindow->height;
	state = STATE_WINDOWSET;


    } else {
	if (setwindow == 0) {
	    if (DEBUG)
		printf("back in SetWindow\n");
	    if (DEBUG)
		printf("New Size: %ix%i\n", aWindow->width,
		       aWindow->height);
	    if (mode == NP_EMBED) {
		embed_width = aWindow->width;
		embed_height = aWindow->height;
	    }
	    setwindow++;

	    if (state != STATE_GETTING_PLAYLIST) {
		state = STATE_GETTING_PLAYLIST;
		if (href != NULL) {
		    if (isMms(href)) {
			snprintf(td->list->url, 1024, "%s", href);
			state = STATE_STARTED_PLAYER;
			td->list->mmsstream = 1;
			SetupPlayer(this, NULL);
		    }
		} else {
		    if (url != NULL) {
			if (isMms(url)) {
			    snprintf(td->list->url, 1024, "%s", url);
			    state = STATE_STARTED_PLAYER;
			    td->list->mmsstream = 1;
			    SetupPlayer(this, NULL);
			}
		    }
		}
	    }
	}
    }

    if (DEBUG) {
	printf("Exiting SetWindow\n");
    }

    return NPERR_NO_ERROR;
}

NPError nsPluginInstance::NewStream(NPMIMEType type, NPStream * stream,
				    NPBool seekable, uint16 * stype)
{
    printf("NewStream %s\n", stream->url);
    if (baseurl == NULL)
	baseurl = getURLBase((char *) stream->url);

    if (mode == NP_FULL) {
	url = strdup(stream->url);
    }

    if ((state <= STATE_GETTING_PLAYLIST) && (controlwindow == 0)) {
	state = STATE_GETTING_PLAYLIST;
	SetupPlayer(this, NULL);
    }
    *stype = NP_NORMAL;

    return NPERR_NO_ERROR;
}


NPError nsPluginInstance::DestroyStream(NPStream * stream, NPError reason)
{
    int playable, all_retrieved, all_above_cache;
    Node *n;

    if (DEBUG)
	printf("NPP_DestroyStream called %i\n URL: %s\n", reason,
	       stream->url);


    if (reason == NPRES_DONE) {
	playable = 0;
	if (strlen(stream->url) >= 1023) {
	    return NPERR_NO_ERROR;
	}
	pthread_mutex_lock(&playlist_mutex);
	n = td->list;
	while (n != NULL) {
	    if (URLcmp(n->url, stream->url) == 0) {
		if (DEBUG)
		    printf("Destroy stream found a URL match\n");
		break;
	    } else {
		if (strstr(stream->url, n->url) != NULL) {
		    break;
		}
	    }
	    n = n->next;
	}

	if (n != NULL) {
	    n->retrieved = 1;
	    if (!isMms(n->url)) {
		if (DEBUG)
		    printf("calling buildPlaylist with filename %s\n",
			   n->fname);
		buildPlaylist(this, n->fname, n);
	    } else {
		n->mmsstream = 1;
	    }

	    // test for all retrieved
	    if (threadsignaled == 0) {
		all_retrieved = 1;
		n = td->list;
		while (n != NULL) {
		    if ((n->retrieved == 0) && (n->play == 1)) {
			all_retrieved = 0;
			break;
		    }
		    n = n->next;
		}
		// if all retrieved signal player
		if (all_retrieved) {
		    if (threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player (retrieved)\n");
			if (autostart) {
			    pthread_cond_signal(&playlist_complete_cond);
			    threadsignaled = 1;
			}
		    }
		}
	    }
	    // if not then see if all not retrieved are above cache,
	    if (threadsignaled == 0) {
		all_above_cache = 1;
		n = td->list;
		while (n != NULL) {
		    if ((n->bytes < n->cachebytes) && (n->play == 1)) {
			all_above_cache = 0;
			break;
		    }
		    n = n->next;
		}
		// if all above cache signal player
		if (all_above_cache) {
		    if (threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player\n");
			if (autostart) {
			    pthread_cond_signal(&playlist_complete_cond);
			    threadsignaled = 1;
			}
		    }
		}
	    }

	    if (threadsignaled == 0) {
		// look for mmsstream
		n = td->list;
		while (n != NULL) {
		    if ((n->mmsstream == 1) && (n->play == 1)) {
			if (threadsignaled == 0) {
			    if (DEBUG)
				printf("signalling player\n");
			    if (autostart) {
				pthread_cond_signal
				    (&playlist_complete_cond);
				threadsignaled = 1;
			    }
			}
			break;
		    }
		    n = n->next;
		}
	    }
	}
	pthread_mutex_unlock(&playlist_mutex);
    }

    return NPERR_NO_ERROR;
}

int32 nsPluginInstance::WriteReady(NPStream * stream)
{
    Node *n;

    if (DESTROYED)
	return -1;

    if (state == STATE_PLAY_CANCELLED)
	return -1;

    if (td == NULL)
	return -1;

    if (strlen(stream->url) >= 1023)
	return -1;

    pthread_mutex_lock(&playlist_mutex);

    if (0)
	printf("WriteReady for %s\n", stream->url);

    n = td->list;
    while (n != NULL) {
	if (strlen(n->url) != 0) {
	    if (URLcmp(n->url, stream->url) == 0) {
		break;
	    } else {
		if (strstr(stream->url, n->url) != NULL) {
		    break;
		}
	    }
	} else {
	    snprintf(n->url, 1024, "%s", stream->url);
	    break;
	}
	n = n->next;
    }

    if (n != NULL) {
	// if we have determined that we are not going to play the file
	// stop downloading it, and clean it up during cleanup
	if (n->cancelled == 1) {
	    n->remove = 1;
	    NPN_DestroyStream(mInstance, stream, NPRES_DONE);
	}
	if (strlen(n->fname) == 0) {
	    snprintf(n->fname, 1024, "%s",
		     tempnam("/tmp", "mplayerplug-inXXXXXX"));
	    if (DEBUG)
		printf("WR tempname: %s\n", n->fname);

	}
	if (n->totalbytes != (int) stream->end)
	    n->totalbytes = stream->end;

	if (n->cachebytes < (long int) (stream->end * cache_percent / 100))
	    n->cachebytes = (long int) (stream->end * cache_percent / 100);

	if (n->cachebytes < (cachesize * 1024))
	    n->cachebytes = cachesize * 1024;

	pthread_mutex_unlock(&playlist_mutex);
	return STREAMBUFSIZE;
    } else {
	if (DEBUG)
	    printf("didn't find the node in the playlist\n %s\n",
		   stream->url);
	n = newNode();
	snprintf(n->url, 1024, "%s", stream->url);
	snprintf(n->fname, 1024, "%s",
		 tempnam("/tmp", "mplayerplug-inXXXXXX"));
	addToEnd(td->list, n);
	if (n->totalbytes != (int) stream->end)
	    n->totalbytes = stream->end;
	pthread_mutex_unlock(&playlist_mutex);
	return STREAMBUFSIZE;
    }

}

int32 nsPluginInstance::Write(NPStream * stream, int32 offset, int32 len,
			      void *buffer)
{
    int ret = 0;
    long int currdownload, maxdownload;
    FILE *fp;
    Node *n, *hrefnode;
    char message[1024];
    char *burl;

    //printf("entering write\n");
    if (DESTROYED)
	return -1;
    if (state == STATE_PLAY_CANCELLED)
	return -1;
    if (td == NULL)
	return -1;

    if (0)
	printf("Write %s : %i : %i\n", stream->url, offset, len);
    if (strlen(stream->url) >= 1023)
	return -1;

    pthread_mutex_lock(&playlist_mutex);

    // find the node that matches the URL and open the file name in it.
    currdownload = 0;
    maxdownload = 0;
    for (n = td->list; n != NULL; n = n->next) {
	//if (URLcmp(n->url, stream->url) == 0) continue;
	//if (strstr(stream->url, n->url) == NULL) continue;
	if (n->play == 0)
	    continue;
	if (n->cancelled)
	    continue;
	currdownload += n->bytes;
	maxdownload += n->totalbytes;
    }

    n = td->list;
    while (n != NULL) {
	if (URLcmp(n->url, stream->url) == 0) {
	    break;
	} else {
	    if (strstr(stream->url, n->url) != NULL) {
		break;
	    }
	}
	n = n->next;
    }

    if (n == NULL) {
	pthread_mutex_unlock(&playlist_mutex);
	return -1;
    } else {
	if (n->cancelled == 1) {
	    pthread_mutex_unlock(&playlist_mutex);
	    return -1;
	}

	if (n->status != STATE_CANCELLED) {
	    fp = fopen(n->fname, "a+");
	    if (fp == NULL) {
		pthread_mutex_unlock(&playlist_mutex);
		return -1;
	    }
	    fseek(fp, offset, SEEK_SET);
	    ret = fwrite(buffer, 1, len, fp);
	    n->bytes = n->bytes + ret;
	    snprintf(message, 1024, "Downloaded %li bytes", n->bytes);
	    if (state < STATE_PLAYING)
		DrawUI(widget, this, message,
		       0,
		       (int) ((currdownload / (maxdownload * 1.0)) * 100));
	    fclose(fp);

	    if (href != NULL) {
		if (hrefrequested == 0) {
		    if (DEBUG)
			printf("requesting href\n");
		    hrefrequested = 1;
		    n = td->list;
		    while (n != NULL) {
			if (URLcmp(n->url, href) == 0) {
			    break;
			} else {
			    if (strstr(stream->url, href) != NULL) {
				break;
			    }
			}
			n = n->next;
		    }
		    if (n == NULL) {
			if (DEBUG)
			    printf("href not in list\n");
			hrefnode = newNode();
			strlcpy(hrefnode->url, href,
				sizeof(hrefnode->url));
			addToEnd(td->list, hrefnode);
		    } else {
			if (DEBUG)
			    printf("href in list\n");
		    }
		    NPN_GetURL(mInstance, href, NULL);
		    pthread_mutex_unlock(&playlist_mutex);
		    return ret;
		}
	    }

	    if (n->status != STATE_DOWNLOADED_ENOUGH) {
		burl = getURLBase(n->url);
		if (burl != NULL) {
		    if (baseurl == NULL) {
			baseurl = burl;
		    } else {
			if (strcmp(baseurl, burl) != 0) {
			    NPN_MemFree(baseurl);
			    baseurl = burl;
			} else {
			    NPN_MemFree(burl);
			}
		    }
		}

		if (isMms(n->url)) {
		    n->mmsstream = 1;
		}
	    }
	    if (n->play == 1) {
		if (n->mmsstream == 1) {
		    if (threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player from write\n");
			if (autostart) {
			    pthread_cond_signal(&playlist_complete_cond);
			    threadsignaled = 1;
			}
		    }
		} else {
		    //if (This->mode != NP_FULL) {
		    if (n->bytes > n->cachebytes) {
			if (threadsignaled == 0) {
			    if (DEBUG)
				printf("signalling player from write\n");
			    if (autostart) {
				pthread_cond_signal
				    (&playlist_complete_cond);
				threadsignaled = 1;
			    }
			}
		    }
		    //}
		}
	    }

	}
	n->status = STATE_DOWNLOADED_ENOUGH;
	pthread_mutex_unlock(&playlist_mutex);
	return ret;
    }


}

// methods called from nsScriptablePeer

void nsPluginInstance::Play()
{
    if (paused == 1) {
	if (DEBUG)
	    printf("sending play\n");
	sendCommand(control, "pause\n");
	paused = 0;
	js_state = JS_STATE_PLAYING;
    }

    if (threadsignaled == 1 && js_state == JS_STATE_UNDEFINED) {

	pthread_create(&player_thread, &thread_attr,
		       playPlaylist, (void *) td);
	// recommended slight pause
	usleep(1);
	pthread_cond_signal(&playlist_complete_cond);
	threadsignaled = 1;
    }

    if ((autostart == 0) && (threadsignaled == 0)) {
	pthread_cond_signal(&playlist_complete_cond);
	threadsignaled = 1;
    }

}

void nsPluginInstance::Pause()
{
    if (paused == 0) {
	if (DEBUG)
	    printf("sending pause\n");
	sendCommand(control, "pause\n");
	paused = 1;
	js_state = JS_STATE_PAUSED;
    }
}

void nsPluginInstance::Stop()
{
    if (DEBUG)
	printf("sending stop\n");
    sendCommand(control, "seek 0 2\npause\n");
    paused = 1;
    js_state = JS_STATE_STOPPED;
}

void nsPluginInstance::Quit()
{
    if (DEBUG)
	printf("sending quit\n");
    if (paused == 1)
	sendCommand(control, "pause\n");
    sendCommand(control, "quit\n");
}

void nsPluginInstance::FastForward()
{
    if (DEBUG)
	printf("sending FastForward\n");
    sendCommand(control, "seek +10 0\n");
}

void nsPluginInstance::FastReverse()
{
    if (DEBUG)
	printf("sending FastReverse\n");
    sendCommand(control, "seek -10 0\n");
}

void nsPluginInstance::GetPlayState(PRInt32 * playstate)
{
    *playstate = js_state;
}

void nsPluginInstance::Seek( double counter )
{
	char command[32];
	FILE *error = fopen("/tmp/mplayerplug-in.log-Seek", "a");
		
	sprintf(command, "seek %5.0f 2", counter);
	fprintf( error, "Seek command: %s\n", command );
	fclose( error );

	sendCommand( control, command );
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// here the plugin is asked by Mozilla to tell if it is scriptable
// we should return a valid interface id and a pointer to
// nsScriptablePeer interface which we should have implemented
// and which should be defined in the corressponding *.xpt file
// in the bin/components folder
NPError nsPluginInstance::GetValue(NPPVariable aVariable, void *aValue)
{
    NPError rv = NPERR_NO_ERROR;

    switch (aVariable) {
    case NPPVpluginScriptableInstance:{
	    // addref happens in getter, so we don't addref here
	    nsIScriptableMplayerPlugin *scriptablePeer =
		getScriptablePeer();
	    if (scriptablePeer) {
		*(nsISupports **) aValue = scriptablePeer;
	    } else
		rv = NPERR_OUT_OF_MEMORY_ERROR;
	}
	break;

    case NPPVpluginScriptableIID:{
	    static nsIID scriptableIID = NS_ISCRIPTABLEMPLAYERPLUGIN_IID;
	    nsIID *ptr = (nsIID *) NPN_MemAlloc(sizeof(nsIID));
	    if (ptr) {
		*ptr = scriptableIID;
		*(nsIID **) aValue = ptr;
	    } else
		rv = NPERR_OUT_OF_MEMORY_ERROR;
	}
	break;

    default:
	break;
    }

    return rv;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// this method will return the scriptable object (and create it if necessary)
nsScriptablePeer *nsPluginInstance::getScriptablePeer()
{
    if (!mScriptablePeer) {
	mScriptablePeer = new nsScriptablePeer(this);
	if (!mScriptablePeer)
	    return NULL;

	NS_ADDREF(mScriptablePeer);
    }
    // add reference for the caller requesting the object
    NS_ADDREF(mScriptablePeer);
    return mScriptablePeer;
}
